import importlib.resources
from typing import Any, Dict, Literal, Optional, Text

import structlog

from rasa.dialogue_understanding.commands.command_syntax_manager import (
    CommandSyntaxVersion,
)
from rasa.dialogue_understanding.generator.single_step.single_step_based_llm_command_generator import (  # noqa: E501
    SingleStepBasedLLMCommandGenerator,
)
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import (
    PROMPT_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
)
from rasa.shared.utils.io import raise_deprecation_warning
from rasa.shared.utils.llm import (
    check_prompt_config_keys_and_warn_if_deprecated,
    get_prompt_template,
)

DEFAULT_COMMAND_PROMPT_TEMPLATE = importlib.resources.read_text(
    "rasa.dialogue_understanding.generator.prompt_templates",
    "command_prompt_template.jinja2",
)


structlogger = structlog.get_logger()


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.COMMAND_GENERATOR,
    ],
    is_trainable=True,
)
class SingleStepLLMCommandGenerator(SingleStepBasedLLMCommandGenerator):
    """A single step LLM-based command generator."""

    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        prompt_template: Optional[Text] = None,
        **kwargs: Any,
    ) -> None:
        raise_deprecation_warning(
            message=(
                "Support for `SingleStepLLMCommandGenerator` will be removed in Rasa "
                "`4.0.0`. Please modify your assistant's configuration to use the "
                "`CompactLLMCommandGenerator` or `SearchReadyLLMCommandGenerator` "
                "instead."
            )
        )
        super().__init__(
            config,
            model_storage,
            resource,
            prompt_template=prompt_template,
            **kwargs,
        )

        # Warn if the prompt config key is used to set the prompt template
        check_prompt_config_keys_and_warn_if_deprecated(
            config, "single_step_llm_command_generator"
        )

    @staticmethod
    def get_component_command_syntax_version() -> CommandSyntaxVersion:
        return CommandSyntaxVersion.v1

    @classmethod
    def _resolve_component_prompt_template(
        cls: Any,
        config: Dict[str, Any],
        prompt_template: Optional[str] = None,
        log_context: Optional[Literal["init", "fingerprint_addon"]] = None,
        log_source_component: Optional[str] = "SingleStepLLMCommandGenerator",
    ) -> Optional[str]:
        """Get the prompt template from the config or the default prompt template."""
        # Case when model is being loaded
        if prompt_template is not None:
            return prompt_template

        # The prompt can be configured in the config via the "prompt" (deprecated) or
        # "prompt_template" properties
        prompt_template_path = (
            config.get(PROMPT_CONFIG_KEY)
            or config.get(PROMPT_TEMPLATE_CONFIG_KEY)
            or None
        )

        # Try to load the template from the given path or fallback to the default for
        # the component
        return get_prompt_template(
            prompt_template_path,
            DEFAULT_COMMAND_PROMPT_TEMPLATE,
            log_source_component=log_source_component,
            log_source_method=log_context,
        )



import asyncio
from typing import Any, Dict, List, Optional

import numpy as np
import structlog

from rasa.shared.constants import (
    HUGGINGFACE_LOCAL_EMBEDDING_CACHING_FOLDER,
)
from rasa.shared.providers._configs.huggingface_local_embedding_client_config import (
    HuggingFaceLocalEmbeddingClientConfig,
)
from rasa.shared.providers.embedding.embedding_response import EmbeddingResponse

structlogger = structlog.get_logger()


class HuggingFaceLocalEmbeddingClient:
    """This client facilitates the computation of embeddings using locally
    stored HuggingFace transformer models. It utilizes the
    `sentence-transformers` package.

    To use, you should have the `sentence_transformers` python package
    installed.

    Parameters:
        model (str): Identifier for the model or path to a directory containing
            model files.
        cache_folder (Optional[str]): Path to the folder for caching models.
            Default is defined by `HUGGINGFACE_LOCAL_EMBEDDING_CACHING_FOLDER`.
        model_kwargs (Optional[Dict[str, Any]]): Additional keyword arguments
            for model configuration.
        encode_kwargs (Optional[Dict[str, Any]]): Parameters to control the
            encoding process.
        multi_process (bool): Flag to enable or disable multiprocessing for
            embedding generation.
        show_progress (bool): Whether to show a progress bar during embedding
            generation.
    """

    def __init__(
        self,
        model: str,
        cache_folder: Optional[str] = str(HUGGINGFACE_LOCAL_EMBEDDING_CACHING_FOLDER),
        model_kwargs: Optional[Dict[str, Any]] = None,
        encode_kwargs: Optional[Dict[str, Any]] = None,
        multi_process: bool = False,
        show_progress: bool = True,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._model_name_or_path = model
        self._multi_process = multi_process
        self._cache_folder = cache_folder
        self._model_kwargs = model_kwargs or {}
        self._encode_kwargs = encode_kwargs or {}
        self._show_progress = show_progress
        self._init_client()

    def _init_client(self) -> None:
        self._validate_if_sentence_transformers_installed()

        import sentence_transformers

        self._client = sentence_transformers.SentenceTransformer(
            model_name_or_path=self._model_name_or_path,
            cache_folder=self._cache_folder,
            **self._model_kwargs,
        )

    @property
    def model(self) -> str:
        return self._model_name_or_path

    @property
    def config(self) -> Dict:
        config = HuggingFaceLocalEmbeddingClientConfig(
            model=self._model_name_or_path,
            cache_folder=self._cache_folder,
            show_progress=self._show_progress,
            model_kwargs=self._model_kwargs,
            encode_kwargs=self._encode_kwargs,
            multi_process=self._multi_process,
        )
        return config.to_dict()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "HuggingFaceLocalEmbeddingClient":
        huggingface_config = HuggingFaceLocalEmbeddingClientConfig.from_dict(config)
        return cls(
            model=huggingface_config.model,
            cache_folder=huggingface_config.cache_folder,
            model_kwargs=huggingface_config.model_kwargs,
            encode_kwargs=huggingface_config.encode_kwargs,
            show_progress=huggingface_config.show_progress,
        )

    @staticmethod
    def _validate_if_sentence_transformers_installed() -> None:
        """In order to use the local embeddings from HuggingFace,
        sentence-transformer must be installed.

        Raises:
            ImportError: If 'sentence-transformers' is not installed.
        """
        try:
            import sentence_transformers  # noqa: F401
        except ImportError as exc:
            message = (
                "Could not import sentence_transformers python package. "
                "Please install it with 'pip install sentence-transformers'."
            )
            structlogger.error(
                "huggingface_local_embedding_client.sentence_transformers_not_installed",
                message=message,
            )
            raise ImportError(message) from exc

    def embed(self, documents: List[str]) -> EmbeddingResponse:
        """Compute doc embeddings using a HuggingFace transformer model.

        Args:
            documents: The list of documents to embed.

        Returns:
            List of embeddings, one for each text.
        """
        self._validate_if_sentence_transformers_installed()
        documents = list(map(lambda x: x.replace("\n", " "), documents))

        if self._multi_process:
            embeddings = self._embed_with_multiprocessing(documents)
        else:
            embeddings = self._embed_without_multiprocessing(documents)

        return EmbeddingResponse(
            data=embeddings.tolist(),
            model=self._model_name_or_path,
        )

    def _embed_without_multiprocessing(self, documents: List[str]) -> np.ndarray:
        embeddings: np.ndarray = self._client.encode(
            documents,
            show_progress_bar=self._show_progress,
            convert_to_numpy=True,
            **self._encode_kwargs,
        )
        return embeddings

    def _embed_with_multiprocessing(self, documents: List[str]) -> np.ndarray:
        from sentence_transformers import SentenceTransformer

        pool = self._client.start_multi_process_pool()
        embeddings = self._client.encode_multi_process(documents, pool)
        SentenceTransformer.stop_multi_process_pool(pool)
        return embeddings

    async def aembed(self, documents: List[str]) -> EmbeddingResponse:
        """Asynchronous Embed search docs.

        Args:
            documents: List of documents to embed.

        Returns:
            List of embeddings.
        """
        loop = asyncio.get_running_loop()
        # Using run_in_executor to execute the synchronous embedding function
        # in a separate thread
        embeddings = await loop.run_in_executor(None, self.embed, documents)
        return embeddings

    def validate_documents(self, documents: List[str]) -> None:
        for doc in documents:
            if not isinstance(doc, str):
                message = "All documents must be strings."
                structlogger.error(
                    message=message,
                    doc=doc,
                )
                raise ValueError(message)
            if not doc.strip():
                message = "Documents cannot be empty or whitespace."
                structlogger.error(
                    message=message,
                    doc=doc,
                )
                raise ValueError(message)

    def validate_client_setup(self, *args, **kwargs) -> None:  # type: ignore
        pass



PRIVACY_CONFIG_SCHEMA = "privacy/privacy_config_schema.json"
REDACTION_CHAR_KEY = "redaction_char"
KEEP_LEFT_KEY = "keep_left"
KEEP_RIGHT_KEY = "keep_right"
DELETION_KEY = "deletion"
ANONYMIZATION_KEY = "anonymization"
TRACKER_STORE_SETTINGS = "tracker_store_settings"
SLOT_KEY = "slot"
TEXT_KEY = "text"
ENTITIES_KEY = "entities"
VALUE_KEY = "value"
ENTITY_LABEL_KEY = "label"

USER_CHAT_INACTIVITY_IN_MINUTES_ENV_VAR_NAME = "USER_CHAT_INACTIVITY_IN_MINUTES"
GLINER_MODEL_PATH_ENV_VAR_NAME = "GLINER_MODEL_PATH"
HUGGINGFACE_CACHE_DIR_ENV_VAR_NAME = "HUGGINGFACE_HUB_CACHE_DIR"

DEFAULT_PII_MODEL = "urchade/gliner_multi_pii-v1"
GLINER_LABELS = [
    "person",
    "organization",
    "company",
    "phone number",
    "address",
    "full address",
    "postcode",
    "zip code",
    "passport number",
    "email",
    "credit card number",
    "social security number",
    "health insurance id number",
    "date of birth",
    "mobile phone number",
    "bank account number",
    "medication",
    "cpf",
    "driver's license number",
    "tax identification number",
    "medical condition",
    "identity card number",
    "national id number",
    "ip address",
    "email address",
    "iban",
    "credit card expiration date",
    "username",
    "health insurance number",
    "registration number",
    "student id number",
    "insurance number",
    "membership number",
    "booking number",
    "landline phone number",
    "blood type",
    "cvv",
    "reservation number",
    "digital signature",
    "social media handle",
    "license plate number",
    "cnpj",
    "postal code",
    "passport_number",
    "serial number",
    "vehicle registration number",
    "fax number",
    "visa number",
    "insurance company",
    "identity document number",
    "transaction number",
    "national health insurance number",
    "cvc",
    "birth certificate number",
    "train ticket number",
    "passport expiration date",
    "social_security_number",
    "personally identifiable information",
    "banking routing number",
    "sort code",
    "routing number",
    "tax number",
    "swift code",
]



import argparse
import asyncio
import logging
import signal
from pathlib import Path
from typing import Iterable, List, Optional, Text, Tuple, Union

import aiohttp
import ruamel.yaml as yaml

import rasa.cli.utils
import rasa.shared.utils.cli
import rasa.shared.utils.io
import rasa.utils.common
import rasa.utils.io
from rasa.cli import SubParsersAction
from rasa.cli.arguments import x as arguments
from rasa.core.available_endpoints import AvailableEndpoints
from rasa.shared.constants import (
    DEFAULT_CREDENTIALS_PATH,
    DEFAULT_ENDPOINTS_PATH,
)
from rasa.shared.utils.yaml import read_config_file

logger = logging.getLogger(__name__)


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all rasa x parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    x_parser_args = {
        "parents": parents,
        "conflict_handler": "resolve",
        "formatter_class": argparse.ArgumentDefaultsHelpFormatter,
    }

    x_parser_args["help"] = (
        "Run a Rasa server in a mode that enables connecting "
        "to Rasa Enterprise as the config endpoint."
    )

    shell_parser = subparsers.add_parser("x", **x_parser_args)
    shell_parser.set_defaults(func=rasa_x)

    arguments.set_x_arguments(shell_parser)


def _rasa_service(
    args: argparse.Namespace,
    endpoints: AvailableEndpoints,
    rasa_x_url: Optional[Text] = None,
    credentials_path: Optional[Text] = None,
) -> None:
    """Starts the Rasa application."""
    from rasa.core.run import serve_application

    # needs separate logging configuration as it is started in its own process
    rasa.utils.common.configure_logging_and_warnings(args.loglevel)
    rasa.utils.io.configure_colored_logging(args.loglevel)

    if not credentials_path:
        credentials_path = _prepare_credentials_for_rasa_x(
            args.credentials, rasa_x_url=rasa_x_url
        )

    serve_application(
        endpoints=endpoints,
        port=args.port,
        credentials=credentials_path,
        cors=args.cors,
        auth_token=args.auth_token,
        enable_api=True,
        jwt_secret=args.jwt_secret,
        jwt_method=args.jwt_method,
        ssl_certificate=args.ssl_certificate,
        ssl_keyfile=args.ssl_keyfile,
        ssl_ca_file=args.ssl_ca_file,
        ssl_password=args.ssl_password,
    )


def _prepare_credentials_for_rasa_x(
    credentials_path: Optional[Text], rasa_x_url: Optional[Text] = None
) -> Text:
    if credentials_path:
        credentials_path = str(
            rasa.cli.utils.get_validated_path(
                credentials_path, "credentials", DEFAULT_CREDENTIALS_PATH, True
            )
        )
        credentials = read_config_file(credentials_path)
    else:
        credentials = {}

    # this makes sure the Rasa X is properly configured no matter what
    if rasa_x_url:
        credentials["rasa"] = {"url": rasa_x_url}
    dumped_credentials = yaml.dump(credentials, default_flow_style=False)
    tmp_credentials = rasa.utils.io.create_temporary_file(dumped_credentials, "yml")

    return tmp_credentials


def rasa_x(args: argparse.Namespace) -> None:
    """Run Rasa with the `x` subcommand."""
    from rasa.cli.utils import signal_handler

    signal.signal(signal.SIGINT, signal_handler)

    if args.production:
        run_in_enterprise_connection_mode(args)
    else:
        rasa.shared.utils.io.raise_warning(
            "Running Rasa X in local mode is no longer supported as Rasa has "
            "stopped supporting the Community Edition (free version) of ‘Rasa X’."
            "For more information please see "
            "https://rasa.com/blog/rasa-x-community-edition-changes/",
            UserWarning,
        )
        exit()


async def _pull_runtime_config_from_server(
    config_endpoint: Optional[Text],
    attempts: int = 60,
    wait_time_between_pulls: float = 5,
    keys: Iterable[Text] = ("endpoints", "credentials"),
) -> List[Text]:
    """Pull runtime config from `config_endpoint`.

    Returns a list of paths to yaml dumps, each containing the contents of one of
    `keys`.
    """
    while attempts:
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(config_endpoint) as resp:
                    if resp.status == 200:
                        rjs = await resp.json()
                        try:
                            return [
                                rasa.utils.io.create_temporary_file(rjs[k])
                                for k in keys
                            ]
                        except KeyError as e:
                            rasa.shared.utils.cli.print_error_and_exit(
                                "Failed to find key '{}' in runtime config. "
                                "Exiting.".format(e)
                            )
                    else:
                        logger.debug(
                            "Failed to get a proper response from remote "
                            "server. Status Code: {}. Response: '{}'"
                            "".format(resp.status, await resp.text())
                        )
        except aiohttp.ClientError as e:
            logger.debug(f"Failed to connect to server. Retrying. {e}")

        await asyncio.sleep(wait_time_between_pulls)
        attempts -= 1

    rasa.shared.utils.cli.print_error_and_exit(
        "Could not fetch runtime config from server at '{}'. Exiting.".format(
            config_endpoint
        )
    )


def run_in_enterprise_connection_mode(args: argparse.Namespace) -> None:
    """Run Rasa in a mode that enables using Rasa X as the config endpoint."""
    from rasa.shared.utils.cli import print_success

    print_success("Starting a Rasa server in Rasa Enterprise connection mode... 🚀")

    credentials_path, endpoints_path = _get_credentials_and_endpoints_paths(args)
    endpoints = AvailableEndpoints.get_instance(endpoints_path)

    _rasa_service(args, endpoints, None, credentials_path)


def _get_credentials_and_endpoints_paths(
    args: argparse.Namespace,
) -> Tuple[Optional[Text], Optional[Text]]:
    config_endpoint = args.config_endpoint
    endpoints_config_path: Optional[Union[Path, Text]]

    if config_endpoint:
        endpoints_config_path, credentials_path = asyncio.run(
            _pull_runtime_config_from_server(config_endpoint)
        )
    else:
        endpoints_config_path = rasa.cli.utils.get_validated_path(
            args.endpoints, "endpoints", DEFAULT_ENDPOINTS_PATH, True
        )
        credentials_path = None

    return (
        credentials_path,
        str(endpoints_config_path) if endpoints_config_path else None,
    )



import copy
import datetime
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional

import structlog

from rasa.privacy.constants import (
    DEFAULT_PII_MODEL,
    ENTITIES_KEY,
    ENTITY_LABEL_KEY,
    GLINER_LABELS,
    GLINER_MODEL_PATH_ENV_VAR_NAME,
    HUGGINGFACE_CACHE_DIR_ENV_VAR_NAME,
    TEXT_KEY,
    VALUE_KEY,
)
from rasa.privacy.privacy_config import AnonymizationMethod, AnonymizationType
from rasa.shared.core.events import BotUttered, Event, SlotSet, UserUttered

structlogger = structlog.get_logger(__name__)


class PrivacyFilter:
    """A class to anonymise sensitive information."""

    def __init__(self, anonymization_rules: Dict[str, AnonymizationMethod]) -> None:
        """Initialise the PrivacyFilter."""
        self.anonymization_rules = anonymization_rules
        self.labels = GLINER_LABELS
        self.model = self._load_gliner_model()

    def anonymize(
        self, events: List[Event], prior_sensitive_slot_events: List[Event]
    ) -> List[Event]:
        """Anonymize sensitive information in the events of the current turn.

        The order of priority for PII detection is:
        - firstly, the slot-based approach i.e. identify any defined slots in
        the anonymization rules that could have been set in this turn and
        anonymise the plaintext slot values in all 3 event types
        (UserUttered, BotUttered, SlotSet)
        - secondly, the GLiNER model based approach i.e. identify any PII entities
         and anonymise the text in UserUttered events or values of
         SlotSet events that fill from_text slots.
        """
        anonymized_events: List[Event] = []
        anonymized_slots = self._anonymize_sensitive_slots(
            (events + prior_sensitive_slot_events)
        )

        for event in events:
            anonymized_event = self._anonymize_event(event, anonymized_slots)
            anonymized_events.append(anonymized_event)

        return anonymized_events

    @staticmethod
    def _load_gliner_model() -> Optional[Any]:
        """Load the GLiNER model for PII detection."""
        local_model_path = os.getenv(GLINER_MODEL_PATH_ENV_VAR_NAME)
        cache_dir_env_value = os.getenv(HUGGINGFACE_CACHE_DIR_ENV_VAR_NAME)
        cache_dir = Path(cache_dir_env_value).resolve() if cache_dir_env_value else None
        model_path = (
            Path(local_model_path).resolve() if local_model_path else DEFAULT_PII_MODEL
        )
        local_files_only = isinstance(model_path, Path) and model_path.exists()

        structlogger.debug(
            "rasa.privacy.privacy_filter.loading_gliner_model",
            local_files_only=local_files_only,
        )

        try:
            from gliner import GLiNER

            return GLiNER.from_pretrained(
                model_path,
                cache_dir=cache_dir,
                local_files_only=local_files_only,
            )
        except ImportError:
            structlogger.warning(
                "rasa.privacy.privacy_filter.gliner_import_error",
                event_info="Optional GLiNER library is not installed. "
                "Please install it if you wish to use additional "
                "PII detection to the slot based approach.",
            )
            return None

    def _anonymize_sensitive_slots(self, events: List[Event]) -> Dict[str, SlotSet]:
        """Identify and anonymize sensitive slot events.

        Returns a dictionary where the keys represent a concatenation of the slot key
        and its original value, and the values are the anonymized SlotSet events.
        """
        sensitive_slots = self._find_sensitive_slots(events)

        if not sensitive_slots:
            structlogger.debug("rasa.privacy.privacy_filter.no_sensitive_slots_found")
            return {}

        anonymized_slots = {}
        for slot in sensitive_slots:
            slot_value = (
                slot.value if isinstance(slot.value, str) else json.dumps(slot.value)
            )
            anonymized_slots[f"{slot.key}:{slot_value}"] = (
                self._anonymize_sensitive_slot_event(slot)
            )

        return anonymized_slots

    def _find_sensitive_slots(self, processed_events: List[Event]) -> List[SlotSet]:
        """Find all slot events that contain sensitive information.

        These sensitive slots are defined in the anonymization rules and
        have a non-empty value.
        """
        return [
            copy.deepcopy(slot_event)
            for slot_event in processed_events
            if isinstance(slot_event, SlotSet)
            and slot_event.key in self.anonymization_rules
            and bool(slot_event.value)
        ]

    def _anonymize_sensitive_slot_event(
        self,
        slot_event: SlotSet,
    ) -> SlotSet:
        """Anonymize the sensitive slot event if it contains sensitive information.

        A sensitive slot event is defined as a SlotSet event that has a key
        in the anonymization rules and a non-empty value.
        """
        slot_value = slot_event.value
        if not bool(slot_value):
            return slot_event

        anonymized_value = self._anonymize_value(slot_event)
        slot_event.value = anonymized_value

        return slot_event

    def _anonymize_event(
        self, event: Event, anonymized_slots: Dict[str, SlotSet]
    ) -> Event:
        if isinstance(event, SlotSet):
            return self._anonymize_slot_event(event, anonymized_slots)
        elif isinstance(event, UserUttered):
            return self._anonymize_user_event(event, anonymized_slots)
        elif isinstance(event, BotUttered):
            return self._anonymize_bot_event(event, anonymized_slots)
        else:
            return event

    def _anonymize_slot_event(
        self,
        event: SlotSet,
        anonymized_slots: Dict[str, SlotSet],
    ) -> SlotSet:
        """Anonymize the slot event if it contains sensitive information."""
        event_value = (
            event.value if isinstance(event.value, str) else json.dumps(event.value)
        )
        # obtain the anonymized slot event, otherwise return the original event
        slot_event = anonymized_slots.get(f"{event.key}:{event_value}", event)

        # apply the edge case anonymization
        slot_value = (
            slot_event.value
            if isinstance(slot_event.value, str)
            else json.dumps(slot_event.value)
        )
        anonymized_value = self._anonymize_edge_cases(slot_value, anonymized_slots)

        slot_event.value = (
            anonymized_value
            if isinstance(slot_event.value, str)
            else json.loads(anonymized_value)
        )
        slot_event.anonymized_at = datetime.datetime.now(datetime.timezone.utc)
        return slot_event

    def _anonymize_user_event(
        self,
        user_event: UserUttered,
        anonymized_slots: Dict[str, SlotSet],
    ) -> UserUttered:
        """Anonymize the user event if it contains sensitive information."""
        if not user_event.text:
            structlogger.debug(
                "rasa.privacy.privacy_filter.user_event_no_text",
            )
            return user_event

        original_parse_data: Dict[str, Any] = (
            copy.deepcopy(user_event.parse_data) if user_event.parse_data else {}
        )
        anonymized_parse_data: Dict[str, Any] = {}

        for key, slot in anonymized_slots.items():
            original_slot_value = key.split(":", 1)[1]
            anonymized_text = self._smart_replace(
                user_event.text, original_slot_value, slot.value
            )
            user_event.text = anonymized_text

            anonymized_parse_data[TEXT_KEY] = anonymized_text
            for entity in original_parse_data.get(ENTITIES_KEY, []):
                entity_value = entity[VALUE_KEY]
                if entity_value == original_slot_value:
                    anonymized_entities: List[Dict[str, Any]] = (
                        anonymized_parse_data.get(ENTITIES_KEY, [])
                    )
                    anonymized_entities.append({**entity, VALUE_KEY: slot.value})
                    anonymized_parse_data[ENTITIES_KEY] = anonymized_entities

        user_event.parse_data = anonymized_parse_data  # type: ignore[assignment]
        user_event.text = self._anonymize_edge_cases(user_event.text, anonymized_slots)
        # cover the edge case anonymization for the parse data text field
        parse_data_text = user_event.parse_data.get(TEXT_KEY, "")
        user_event.parse_data[TEXT_KEY] = self._anonymize_edge_cases(  # type: ignore[literal-required]
            parse_data_text, anonymized_slots
        )

        user_event.anonymized_at = datetime.datetime.now(datetime.timezone.utc)

        return user_event

    def _anonymize_bot_event(
        self,
        bot_event: BotUttered,
        anonymized_slots: Dict[str, SlotSet],
    ) -> BotUttered:
        """Anonymize the bot event if it contains sensitive information."""
        if not bot_event.text:
            structlogger.debug(
                "rasa.privacy.privacy_filter.bot_event_no_text",
            )
            return bot_event

        for key, slot in anonymized_slots.items():
            original_slot_value = key.split(":", 1)[1]
            anonymized_text = self._smart_replace(
                bot_event.text, original_slot_value, slot.value
            )
            bot_event.text = anonymized_text

        bot_event.text = self._anonymize_edge_cases(bot_event.text, anonymized_slots)
        bot_event.anonymized_at = datetime.datetime.now(datetime.timezone.utc)
        return bot_event

    def _anonymize_value(self, slot: SlotSet) -> str:
        """Anonymize the given slot value using the specified anonymization method."""
        slot_name = slot.key
        slot_value = slot.value
        anonymization_method = self.anonymization_rules[slot_name]

        if anonymization_method.method_type == AnonymizationType.REDACT:
            return self._redact(slot_value, anonymization_method)

        if anonymization_method.method_type == AnonymizationType.MASK:
            return self._mask(slot_name)

        # we won't reach this case, because the json schema specifies
        # the allowed methods, this is to satisfy the type checker
        return ""

    @staticmethod
    def _redact(slot_value: Any, anonymization_method: AnonymizationMethod) -> str:
        """Redact the given slot value using the specified anonymization method."""
        if anonymization_method.keep_left is not None:
            left_part = slot_value[: anonymization_method.keep_left]
        else:
            left_part = ""

        if anonymization_method.keep_right is not None:
            right_part = slot_value[-anonymization_method.keep_right :]
        else:
            right_part = ""

        return (
            left_part
            + anonymization_method.redaction_char
            * (len(slot_value) - len(left_part) - len(right_part))
            + right_part
        )

    @staticmethod
    def _mask(slot_name: str) -> str:
        """Mask the given slot value using the slot name."""
        return f"[{slot_name.upper()}]"

    @staticmethod
    def _strip_square_brackets(string: str) -> str:
        """Strip square brackets from the start and end of the string if present."""
        if len(string) >= 2 and string[0] == "[" and string[-1] == "]":
            return string[1:-1]
        return string

    def _anonymize_edge_cases(
        self, text: str, anonymized_slots: Dict[str, SlotSet]
    ) -> str:
        """Anonymize edge cases in the text using GLiNER model.

        This method is used to detect PII entities in the text that are not
        covered by the slot-based anonymization rules. For example, when
        the user message contains PII entities that are not defined as slots,
        or when the slot is filled from a text input that could contain multiple
        PII entities, such as a from_text slot.

        This method uses the GLiNER model to predict entities in the text
        and replaces them with masked values.
        If the GLiNER model is not loaded, it will skip this step and return
        the original text.
        """
        if self.model is None:
            structlogger.debug(
                "rasa.privacy.privacy_filter.gliner_model_not_loaded",
                event_info="GLiNER model is not loaded, skipping PII detection.",
            )
            return text

        entities = self.model.predict_entities(text, self.labels, threshold=0.85)

        all_anonymized_slot_values = [
            self._strip_square_brackets(str(slot.value))
            for slot in anonymized_slots.values()
        ]

        for entity in entities:
            structlogger.debug(
                "rasa.privacy.privacy_filter.pii_entity_found",
                entity=entity[ENTITY_LABEL_KEY],
            )

            entity_value = entity[TEXT_KEY]

            if entity_value in all_anonymized_slot_values:
                # the entity that was found is already anonymized,
                # we shouldn't override the already anonymized value
                # with a masked value
                structlogger.debug(
                    "rasa.privacy.privacy_filter.pii_entity_already_anonymized",
                    entity=entity[ENTITY_LABEL_KEY],
                )
                continue

            text = text.replace(entity_value, self._mask(entity[ENTITY_LABEL_KEY]))

        return text

    @staticmethod
    def _smart_replace(text: str, original_value: str, replacement: str) -> str:
        """Replace original_value with replacement in text.

        This method performs a string replacement in the text,
        with special handling for floats.
        If original_value is a float string like "24.0",
        also tries replacing the integer version "24".

        Args:
            text (str): The text to perform replacements on
            original_value (str): The value to replace
            replacement (str): The replacement value

        Returns:
            str: The text with replacements applied
        """
        # First try the original replacement
        result = text.replace(original_value, replacement)
        if text != result:
            return result

        # If replacement didn't happen and it's a float,
        # try replacing the integer version
        if "." in original_value:
            try:
                float_val = float(original_value)
                if float_val.is_integer():
                    int_version = str(int(float_val))
                    result = result.replace(int_version, replacement)
            except ValueError:
                structlogger.warning(
                    "rasa.privacy.privacy_filter.smart_replace_float_error",
                    event_info="Unable to anonymize float value.",
                )

        return result



# this builtin is needed so we can overwrite in test
import asyncio
import json
import logging
import os
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    List,
    Optional,
    Text,
    overload,
)

import aiohttp
import questionary
from aiohttp import ClientTimeout
from prompt_toolkit.styles import Style

import rasa.shared.utils.cli
import rasa.shared.utils.io
from rasa.cli import utils as cli_utils
from rasa.core import utils
from rasa.core.channels.rest import RestInput
from rasa.core.constants import DEFAULT_SERVER_URL, DEFAULT_STREAM_READING_TIMEOUT
from rasa.shared.constants import INTENT_MESSAGE_PREFIX
from rasa.shared.utils.io import DEFAULT_ENCODING

logger = logging.getLogger(__name__)

STREAM_READING_TIMEOUT_ENV = "RASA_SHELL_STREAM_READING_TIMEOUT_IN_SECONDS"


def print_buttons(
    message: Dict[Text, Any],
    is_latest_message: bool = False,
    color: Text = rasa.shared.utils.io.bcolors.OKBLUE,
) -> Optional[questionary.Question]:
    """Create CLI buttons from message data."""
    if is_latest_message:
        choices = cli_utils.button_choices_from_message_data(
            message, allow_free_text_input=True
        )
        question = questionary.select(
            message.get("text"),
            choices,
            style=Style([("qmark", "#6d91d3"), ("", "#6d91d3"), ("answer", "#b373d6")]),
        )
        return question
    else:
        rasa.shared.utils.cli.print_color("Buttons:", color=color)
        for idx, button in enumerate(message.get("buttons")):
            rasa.shared.utils.cli.print_color(
                cli_utils.button_to_string(button, idx), color=color
            )
        return None


def _print_bot_output(
    message: Dict[Text, Any],
    is_latest_message: bool = False,
    color: Text = rasa.shared.utils.io.bcolors.OKBLUE,
) -> Optional[questionary.Question]:
    if "buttons" in message:
        question = print_buttons(message, is_latest_message, color)
        if question:
            return question

    if "text" in message:
        rasa.shared.utils.cli.print_color(message["text"], color=color)

    if "image" in message:
        rasa.shared.utils.cli.print_color("Image: " + message["image"], color=color)

    if "attachment" in message:
        rasa.shared.utils.cli.print_color(
            "Attachment: " + message["attachment"], color=color
        )

    if "elements" in message:
        rasa.shared.utils.cli.print_color("Elements:", color=color)
        for idx, element in enumerate(message["elements"]):
            rasa.shared.utils.cli.print_color(
                cli_utils.element_to_string(element, idx), color=color
            )

    if "quick_replies" in message:
        rasa.shared.utils.cli.print_color("Quick Replies:", color=color)
        for idx, element in enumerate(message["quick_replies"]):
            rasa.shared.utils.cli.print_color(
                cli_utils.button_to_string(element, idx), color=color
            )

    if "custom" in message:
        rasa.shared.utils.cli.print_color("Custom json:", color=color)
        rasa.shared.utils.cli.print_color(
            rasa.shared.utils.io.json_to_string(message["custom"]), color=color
        )

    return None


@overload
async def _get_user_input(previous_response: None) -> Text: ...


@overload
async def _get_user_input(previous_response: Dict[str, Any]) -> Optional[Text]: ...


async def _get_user_input(
    previous_response: Optional[Dict[str, Any]],
) -> Optional[Text]:
    button_response = None
    if previous_response is not None:
        button_response = _print_bot_output(previous_response, is_latest_message=True)

    if button_response is not None:
        response = await cli_utils.payload_from_button_question(button_response)
        if response == cli_utils.FREE_TEXT_INPUT_PROMPT:
            # Re-prompt user with a free text input
            response = await _get_user_input(None)
    else:
        question = questionary.text(
            "",
            qmark="Your input ->",
            style=Style([("qmark", "#b373d6"), ("", "#b373d6")]),
        )
        response = await question.ask_async()
    return response.strip() if response is not None else None


async def send_message_receive_block(
    server_url: Text, auth_token: Text, sender_id: Text, message: Text
) -> List[Dict[Text, Any]]:
    """Posts message and returns response."""
    payload = {"sender": sender_id, "message": message}

    url = f"{server_url}/webhooks/rest/webhook?token={auth_token}"
    async with aiohttp.ClientSession() as session:
        async with session.post(url, json=payload, raise_for_status=True) as resp:
            return await resp.json()


async def _send_message_receive_stream(
    server_url: Text,
    auth_token: Text,
    sender_id: Text,
    message: Text,
    request_timeout: Optional[int] = None,
) -> AsyncGenerator[Dict[Text, Any], None]:
    payload = {"sender": sender_id, "message": message}

    url = f"{server_url}/webhooks/rest/webhook?stream=true&token={auth_token}"

    # Define timeout to not keep reading in case the server crashed in between
    timeout = _get_stream_reading_timeout(request_timeout)

    async with aiohttp.ClientSession(timeout=timeout) as session:
        async with session.post(url, json=payload, raise_for_status=True) as resp:
            async for line in resp.content:
                if line:
                    yield json.loads(line.decode(DEFAULT_ENCODING))


def _get_stream_reading_timeout(request_timeout: Optional[int] = None) -> ClientTimeout:
    """Define the ClientTimeout with fallbacks.

    First use the `request_timeout` function parameter if available, this comes from the
    `--request-timeout` command line argument.
    If that fails fallback `STREAM_READING_TIMEOUT_ENV`, a commandline argument.
    Lastly fallback to `DEFAULT_STREAM_READING_TIMEOUT` from rasa.core.constants
    """
    timeout_str = (
        request_timeout
        if request_timeout is not None
        else os.environ.get(STREAM_READING_TIMEOUT_ENV, DEFAULT_STREAM_READING_TIMEOUT)
    )

    return ClientTimeout(int(timeout_str))


async def record_messages(
    sender_id: Text,
    server_url: Text = DEFAULT_SERVER_URL,
    auth_token: Text = "",
    max_message_limit: Optional[int] = None,
    use_response_stream: bool = True,
    request_timeout: Optional[int] = None,
) -> int:
    """Read messages from the command line and print bot responses."""
    exit_text = INTENT_MESSAGE_PREFIX + "stop"

    rasa.shared.utils.cli.print_success(
        "Bot loaded. Type a message and press enter (use '{}' to exit): ".format(
            exit_text
        )
    )

    num_messages = 0
    previous_response = None
    await asyncio.sleep(0.5)  # Wait for server to start
    while not utils.is_limit_reached(num_messages, max_message_limit):
        text = await _get_user_input(previous_response)

        if text == exit_text or text is None:
            break

        if use_response_stream:
            bot_responses_stream = _send_message_receive_stream(
                server_url, auth_token, sender_id, text, request_timeout=request_timeout
            )
            previous_response = None
            async for response in bot_responses_stream:
                if previous_response is not None:
                    _print_bot_output(previous_response)
                previous_response = response
        else:
            bot_responses = await send_message_receive_block(
                server_url, auth_token, sender_id, text
            )
            previous_response = None
            for response in bot_responses:
                if previous_response is not None:
                    _print_bot_output(previous_response)
                previous_response = response

        num_messages += 1
        await asyncio.sleep(0)  # Yield event loop for others coroutines
    return num_messages


class CmdlineInput(RestInput):
    @classmethod
    def name(cls) -> Text:
        return "cmdline"

    def url_prefix(self) -> Text:
        return RestInput.name()



from typing import TYPE_CHECKING, List, Optional

import structlog

from rasa.core.brokers.broker import EventBroker
from rasa.utils.endpoints import EndpointConfig

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop

structlogger = structlog.get_logger(__name__)


async def create_event_brokers(
    event_broker_endpoint: Optional[EndpointConfig],
    event_loop: Optional["AbstractEventLoop"] = None,
) -> List[EventBroker]:
    """Create EventBroker objects for each anonymization topic or queue."""
    if event_broker_endpoint is None or event_broker_endpoint.type is None:
        structlogger.debug(
            "rasa.privacy_filtering.create_event_broker.no_event_broker_type",
        )
        return []

    if event_broker_endpoint.type == "kafka":
        event_collection = event_broker_endpoint.kwargs.get("anonymization_topics", [])
        event_collection_type = "topic"
    elif event_broker_endpoint.type == "pika":
        event_collection = event_broker_endpoint.kwargs.get("anonymization_queues", [])
        event_collection_type = "queues"
    else:
        structlogger.debug(
            "rasa.privacy_filtering.create_event_broker.unsupported_event_broker_type",
            event_broker_type=event_broker_endpoint.type,
        )
        return []

    if not event_collection:
        structlogger.debug(
            f"rasa.privacy_filtering.create_event_broker.no_anonymization_{event_collection_type}",
            event_collection_type=event_collection_type,
        )
        return []

    return await _create_event_brokers(
        event_broker_endpoint, event_collection, event_collection_type, event_loop
    )


async def _create_event_brokers(
    event_broker_endpoint: EndpointConfig,
    event_collection: List[str],
    event_collection_type: str,
    event_loop: Optional["AbstractEventLoop"] = None,
) -> List[EventBroker]:
    """Create event brokers."""
    event_brokers = []
    for item in event_collection:
        event_broker_endpoint.kwargs[event_collection_type] = (
            item if event_collection_type == "topic" else [item]
        )
        structlogger.debug(
            "rasa.privacy_filtering.create_event_broker",
            event_info=f"Setting anonymized event streaming to '{item}'.",
        )

        event_broker = await EventBroker.create(event_broker_endpoint, event_loop)
        if event_broker is None:
            structlogger.debug(
                "rasa.privacy_filtering.create_event_broker.no_event_broker_created",
                event_info=f"No event broker created for publishing to '{item}'.",
            )
            continue

        event_brokers.append(event_broker)

    return event_brokers



from typing import Any, Dict, List, Optional, Text

from rasa.engine.language import Language
from rasa.shared.core.flows.constants import KEY_TRANSLATION


def get_translated_text(
    text: Optional[Text],
    translation: Dict[Text, Any],
    language: Optional[Language] = None,
) -> Optional[Text]:
    """Get the translated text from the message.

    Args:
        text: The default text to use if no translation is found.
        translation: The translations for the text.
        language: The language to use for the translation.

    Returns:
        The translated text if found, otherwise the default text.
    """
    language_code = language.code if language else None
    return translation.get(language_code, text)


def has_translation(
    message: Dict[Text, Any], language: Optional[Language] = None
) -> bool:
    """Check if the message has a translation for the given language."""
    language_code = language.code if language else None
    return language_code in message.get(KEY_TRANSLATION, {})


def get_translated_buttons(
    buttons: Optional[List[Dict[Text, Any]]], language: Optional[Language] = None
) -> Optional[List[Dict[Text, Any]]]:
    """Get the translated buttons from the message.

    Args:
        buttons: The default buttons to use if no translation is found.
        language: The language to use for the translation.

    Returns:
        The translated buttons if found; otherwise, the default buttons.
    """
    if buttons is None:
        return None

    language_code = language.code if language else None
    translated_buttons = []
    for button in buttons:
        translation = button.get(KEY_TRANSLATION, {})
        language_translation = translation.get(language_code, {})

        # Maintain the original key order to ensure
        # accurate comparisons of BotUtter events.
        translated_button = {
            key: language_translation.get(key, button.get(key))
            for key, value in button.items()
            if key != KEY_TRANSLATION
        }
        translated_buttons.append(translated_button)
    return translated_buttons



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Text, Union

import structlog
from jinja2 import Template
from pypred import Predicate

from rasa.core.actions.action import Action, create_bot_utterance
from rasa.core.utils import add_bot_utterance_metadata
from rasa.dialogue_understanding.patterns.collect_information import (
    CollectInformationPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.validate_slot import (
    ValidateSlotPatternFlowStackFrame,
)
from rasa.shared.core.constants import ACTION_RUN_SLOT_REJECTIONS_NAME
from rasa.shared.core.events import Event, SlotSet
from rasa.shared.core.slots import (
    BooleanSlot,
    CategoricalSlot,
    FloatSlot,
    Slot,
    SlotRejection,
)

if TYPE_CHECKING:
    from rasa.core.channels.channel import OutputChannel
    from rasa.core.nlg import NaturalLanguageGenerator
    from rasa.shared.core.domain import Domain
    from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()

UTTERANCE_NOT_DEFINED_BY_USER = "TO.UPDATE.VALUE.BUT.MISSING.UTTERANCE"


def utterance_for_slot_type(slot: Slot) -> Optional[str]:
    """Return the utterance to use for the slot type."""
    if isinstance(slot, BooleanSlot):
        return "utter_boolean_slot_rejection"
    elif isinstance(slot, FloatSlot):
        return "utter_float_slot_rejection"
    elif isinstance(slot, CategoricalSlot):
        return "utter_categorical_slot_rejection"
    return None


def coerce_slot_value(
    slot_value: str, slot_name: str, tracker: "DialogueStateTracker"
) -> Union[str, bool, float, None]:
    """Coerce the slot value to the correct type.

    Tries to coerce the slot value to the correct type. If the
    conversion fails, `None` is returned.

    Args:
        slot_value: The value to coerce.
        slot_name: The name of the slot.
        tracker: The tracker containing the current state of the conversation.

    Returns:
        The coerced value or `None` if the conversion failed.
    """
    if slot_name not in tracker.slots:
        return slot_value

    slot = tracker.slots[slot_name]

    if not slot.is_valid_value(slot_value):
        structlogger.debug(
            "run.rejection.slot_value_not_valid",
            rejection=slot_value,
        )
        return None

    return slot.coerce_value(slot_value)


def run_rejections(
    slot_value: Union[str, bool, float, None],
    slot_name: str,
    rejections: List[SlotRejection],
) -> Optional[str]:
    """Run the predicate checks under rejections."""
    violation = False
    internal_error = False
    current_context = {"slots": {slot_name: slot_value}}

    structlogger.debug("run.predicate.context", context=current_context)
    document = current_context.copy()

    for rejection in rejections:
        condition = rejection.if_
        utterance = rejection.utter

        try:
            rendered_template = Template(condition).render(current_context)
            predicate = Predicate(rendered_template)
            violation = predicate.evaluate(document)
            structlogger.debug(
                "run.predicate.result",
                predicate=predicate.description(),
                violation=violation,
            )
        except (TypeError, Exception) as e:
            structlogger.error(
                "run.predicate.error",
                predicate=condition,
                error=str(e),
            )
            violation = True
            internal_error = True

        if violation:
            break

    if not violation:
        return None
    if internal_error:
        utterance = "utter_internal_error_rasa"
    if not isinstance(utterance, str):
        utterance = UTTERANCE_NOT_DEFINED_BY_USER
    return utterance


class ActionRunSlotRejections(Action):
    """Action which evaluates the predicate checks under rejections."""

    def name(self) -> Text:
        """Return the name of the action."""
        return ACTION_RUN_SLOT_REJECTIONS_NAME

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Run the predicate checks."""
        utterance = None
        top_frame = tracker.stack.top()
        if not isinstance(
            top_frame,
            (
                CollectInformationPatternFlowStackFrame,
                ValidateSlotPatternFlowStackFrame,
            ),
        ):
            return []
        elif isinstance(top_frame, CollectInformationPatternFlowStackFrame):
            slot_name = top_frame.collect
        elif isinstance(top_frame, ValidateSlotPatternFlowStackFrame):
            slot_name = top_frame.validate

        slot_instance = tracker.slots.get(slot_name)
        if slot_instance and not slot_instance.has_been_set:
            # this is the first time the assistant asks for the slot value,
            # therefore we skip the predicate validation because the slot
            # value has not been provided
            structlogger.debug(
                "first.collect.slot.not.set",
                slot_name=slot_name,
                slot_value=slot_instance.value,
            )
            return []

        slot_value = tracker.get_slot(slot_name)

        if slot_instance and slot_instance.has_been_set and slot_value is None:
            return []

        typed_slot_value = coerce_slot_value(slot_value, slot_name, tracker)
        if typed_slot_value is None:
            # the slot value could not be coerced to the correct type
            utterance = utterance_for_slot_type(slot_instance)
        elif top_frame.rejections:
            # run the predicate checks under rejections
            utterance = run_rejections(
                typed_slot_value, slot_name, top_frame.rejections
            )

        events: List[Event] = []
        if utterance:
            # the slot value has been rejected
            events.append(SlotSet(slot_name, None))
        elif slot_value != typed_slot_value or type(slot_value) != type(
            typed_slot_value
        ):
            # the slot value has been coerced to the correct type
            return [SlotSet(slot_name, typed_slot_value)]
        elif slot_value == typed_slot_value:
            # the slot value has not changed and no utterance present
            return []

        if utterance == UTTERANCE_NOT_DEFINED_BY_USER:
            structlogger.error(
                "run.rejection.missing.utter",
                utterance=None,
            )
            return events

        message = await nlg.generate(
            utterance,
            tracker,
            output_channel.name(),
            value=slot_value,
        )

        if message is None:
            structlogger.error(
                "run.rejection.failed.finding.utter",
                utterance=utterance,
            )
        else:
            message = add_bot_utterance_metadata(
                message, utterance, nlg, domain, tracker
            )
            events.append(create_bot_utterance(message))

        return events



import json
import logging
import typing
from copy import deepcopy
from typing import Any, Awaitable, Callable, Dict, List, Optional, Text

from sanic import Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse

from rasa.core.channels.channel import InputChannel, OutputChannel, UserMessage
from rasa.shared.constants import INTENT_MESSAGE_PREFIX
from rasa.shared.core.constants import USER_INTENT_RESTART
from rasa.shared.exceptions import RasaException

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from aiogram.types import Message, Update


class TelegramOutput(OutputChannel):
    """Output channel for Telegram."""

    # skipcq: PYL-W0236
    @classmethod
    def name(cls) -> Text:
        return "telegram"

    def __init__(self, access_token: Optional[Text]) -> None:
        try:
            from aiogram import Bot

            self.bot = Bot(access_token)
        except ImportError:
            raise ImportError(
                "To use the Telegram channel, please install the aiogram package "
                "with 'pip install aiogram'"
            )

    async def send_text_message(
        self, recipient_id: Text, text: Text, **kwargs: Any
    ) -> None:
        """Sends text message."""
        for message_part in text.strip().split("\n\n"):
            await self.bot.send_message(recipient_id, message_part)

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image."""
        await self.bot.send_photo(recipient_id, image)

    async def send_text_with_buttons(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        button_type: Optional[Text] = "inline",
        **kwargs: Any,
    ) -> None:
        """Sends a message with keyboard.

        For more information: https://core.telegram.org/bots#keyboards

        :button_type inline: horizontal inline keyboard

        :button_type vertical: vertical inline keyboard

        :button_type reply: reply keyboard
        """
        from aiogram.types import InlineKeyboardButton, KeyboardButton
        from aiogram.utils.keyboard import (
            InlineKeyboardBuilder,
            KeyboardBuilder,
            ReplyKeyboardBuilder,
        )

        if button_type == "inline":
            reply_markup_builder: "KeyboardBuilder" = InlineKeyboardBuilder()
            button_list = [
                InlineKeyboardButton(text=s["title"], callback_data=s["payload"])
                for s in buttons
            ]
            reply_markup_builder.row(*button_list)
            reply_markup = reply_markup_builder.as_markup()

        elif button_type == "vertical":
            reply_markup_builder = InlineKeyboardBuilder()
            [
                reply_markup_builder.row(
                    InlineKeyboardButton(text=s["title"], callback_data=s["payload"])
                )
                for s in buttons
            ]
            reply_markup = reply_markup_builder.as_markup()

        elif button_type == "reply":
            reply_markup_builder = ReplyKeyboardBuilder()

            for idx, button in enumerate(buttons):
                if isinstance(button, list):
                    reply_markup_builder.add(
                        *[KeyboardButton(text=s["title"]) for s in button]
                    )
                else:
                    reply_markup_builder.add(KeyboardButton(text=button["title"]))
            reply_markup = reply_markup_builder.as_markup(
                resize_keyboard=False, one_time_keyboard=True
            )
        else:
            logger.error(
                "Trying to send text with buttons for unknown button type {}".format(
                    button_type
                )
            )
            return

        await self.bot.send_message(recipient_id, text, reply_markup=reply_markup)

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Sends a message with a custom json payload."""
        json_message = deepcopy(json_message)

        recipient_id = json_message.pop("chat_id", recipient_id)

        send_functions = {
            ("text",): "send_message",
            ("photo",): "send_photo",
            ("audio",): "send_audio",
            ("document",): "send_document",
            ("sticker",): "send_sticker",
            ("video",): "send_video",
            ("video_note",): "send_video_note",
            ("animation",): "send_animation",
            ("voice",): "send_voice",
            ("media",): "send_media_group",
            ("latitude", "longitude", "title", "address"): "send_venue",
            ("latitude", "longitude"): "send_location",
            ("phone_number", "first_name"): "send_contact",
            ("game_short_name",): "send_game",
            ("action",): "send_chat_action",
            (
                "title",
                "decription",
                "payload",
                "provider_token",
                "start_parameter",
                "currency",
                "prices",
            ): "send_invoice",
        }

        for params in send_functions.keys():
            if all(json_message.get(p) is not None for p in params):
                args = [json_message.pop(p) for p in params]
                api_call = getattr(self.bot, send_functions[params])
                await api_call(recipient_id, *args, **json_message)

    async def get_me(self) -> Any:
        """Get information about the bot itself."""
        return await self.bot.get_me()

    async def set_webhook(self, url: Text) -> None:
        """Set the webhook URL for telegram."""
        await self.bot.set_webhook(url=url)


class TelegramInput(InputChannel):
    """Telegram input channel."""

    @classmethod
    def name(cls) -> Text:
        return "telegram"

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if not credentials:
            cls.raise_missing_credentials_exception()

        return cls(
            credentials.get("access_token"),
            credentials.get("verify"),
            credentials.get("webhook_url"),
        )

    def __init__(
        self,
        access_token: Optional[Text],
        verify: Optional[Text],
        webhook_url: Optional[Text],
        debug_mode: bool = True,
    ) -> None:
        self.access_token = access_token
        self.verify = verify
        self.webhook_url = webhook_url
        self.debug_mode = debug_mode

    @staticmethod
    def _is_location(message: "Message") -> bool:
        return message.location is not None

    @staticmethod
    def _is_user_message(message: "Message") -> bool:
        return message.text is not None

    @staticmethod
    def _is_edited_message(message: "Update") -> bool:
        return message.edited_message is not None

    @staticmethod
    def _is_button(message: "Update") -> bool:
        return message.callback_query is not None

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        telegram_webhook = Blueprint("telegram_webhook", __name__)
        out_channel = self.get_output_channel()

        @telegram_webhook.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @telegram_webhook.route("/set_webhook", methods=["GET", "POST"])
        async def set_webhook(_: Request) -> HTTPResponse:
            try:
                await self.set_webhook(out_channel)
            except RasaException as exc:
                logger.error(exc)
                return response.text(str(exc))

            logger.info("Webhook Setup Successful")
            return response.text("Webhook setup successful")

        @telegram_webhook.route("/webhook", methods=["GET", "POST"])
        async def message(request: Request) -> Any:
            from aiogram.types import Update

            if request.method == "POST":
                request_dict = request.json
                if isinstance(request_dict, Text):
                    request_dict = json.loads(request_dict)
                update = Update(**request_dict)
                credentials = await out_channel.get_me()
                if not credentials.username == self.verify:
                    logger.debug("Invalid access token, check it matches Telegram")
                    return response.text("failed")

                if self._is_button(update):
                    msg = (
                        update.callback_query.message
                        if update.callback_query is not None
                        else None
                    )
                    text = (
                        update.callback_query.data
                        if update.callback_query is not None
                        else None
                    )
                elif self._is_edited_message(update):
                    msg = update.edited_message
                    text = (
                        update.edited_message.text
                        if update.edited_message is not None
                        else None
                    )
                else:
                    msg = update.message
                    if self._is_user_message(msg):
                        text = (
                            msg.text.replace("/bot", "")
                            if msg is not None and msg.text is not None
                            else None
                        )
                    elif self._is_location(msg):
                        text = (
                            '{{"lng":{0}, "lat":{1}}}'.format(
                                msg.location.longitude, msg.location.latitude
                            )
                            if msg is not None and msg.location is not None
                            else None
                        )
                    else:
                        return response.text("success")

                metadata = self.get_metadata(request)
                try:
                    if msg is not None:
                        sender_id = str(msg.chat.id)
                    else:
                        raise ValueError("No message found.")

                    if text == (INTENT_MESSAGE_PREFIX + USER_INTENT_RESTART):
                        await on_new_message(
                            UserMessage(
                                text,
                                out_channel,
                                sender_id,
                                input_channel=self.name(),
                                metadata=metadata,
                            )
                        )
                        await on_new_message(
                            UserMessage(
                                "/start",
                                out_channel,
                                sender_id,
                                input_channel=self.name(),
                                metadata=metadata,
                            )
                        )
                    else:
                        await on_new_message(
                            UserMessage(
                                text,
                                out_channel,
                                sender_id,
                                input_channel=self.name(),
                                metadata=metadata,
                            )
                        )
                except Exception as e:
                    logger.error(f"Exception when trying to handle message.{e}")
                    logger.debug(e, exc_info=True)
                    if self.debug_mode:
                        raise
                    pass

                return response.text("success")

        return telegram_webhook

    def get_output_channel(self) -> TelegramOutput:
        """Loads the telegram channel."""
        return TelegramOutput(self.access_token)

    async def set_webhook(self, channel: TelegramOutput) -> None:
        from aiogram.exceptions import TelegramAPIError

        try:
            await channel.set_webhook(url=self.webhook_url)
        except TelegramAPIError as error:
            raise RasaException(
                "Failed to set channel webhook: " + str(error)
            ) from error



import inspect
import logging
from typing import Any, Dict, Text, Type

import importlib_metadata
from typing_extensions import Protocol, runtime_checkable

import rasa.shared.utils.io
import rasa.utils.common
from rasa.engine.graph import GraphComponent

logger = logging.getLogger(__name__)

import_name_to_package_map = {"sklearn": "scikit_learn"}


@runtime_checkable
class Fingerprintable(Protocol):
    """Interface that enforces training data can be fingerprinted."""

    def fingerprint(self) -> Text:
        """Returns a unique stable fingerprint of the data."""
        ...


def calculate_fingerprint_key(
    graph_component_class: Type[GraphComponent],
    config: Dict[Text, Any],
    inputs: Dict[Text, Fingerprintable],
) -> Text:
    """Calculates a fingerprint key that uniquely represents a single node's execution.

    Args:
        graph_component_class: The graph component class.
        config: The component config.
        inputs: The inputs as a mapping of parent node name to input value.

    Returns:
        The fingerprint key.
    """
    dependency_versions = {
        package: importlib_metadata.version(
            import_name_to_package_map.get(package, package)
        )
        for package in graph_component_class.required_packages()
    }
    fingerprint_data = {
        "node_name": rasa.utils.common.module_path_from_class(graph_component_class),
        "component_implementation": inspect.getsource(graph_component_class),
        "config": config,
        "inputs": inputs,
        "dependency_versions": dependency_versions,
    }

    fingerprint_addon = graph_component_class.fingerprint_addon(config)
    if fingerprint_addon is not None:
        fingerprint_data["addon"] = fingerprint_addon

    fingerprint_key = rasa.shared.utils.io.deep_container_fingerprint(fingerprint_data)

    logger.debug(
        f"Calculated fingerprint_key '{fingerprint_key}' for class "
        f"'{graph_component_class.__name__}'."
    )

    return fingerprint_key



from __future__ import annotations

import audioop
import base64
import json
import uuid
from typing import Any, Awaitable, Callable, Dict, Optional, Tuple

import structlog
from sanic import (  # type: ignore[attr-defined]
    Blueprint,
    HTTPResponse,
    Request,
    Websocket,
    response,
)

from rasa.core.channels import UserMessage
from rasa.core.channels.voice_ready.utils import CallParameters
from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes
from rasa.core.channels.voice_stream.call_state import call_state
from rasa.core.channels.voice_stream.tts.tts_engine import TTSEngine
from rasa.core.channels.voice_stream.util import repack_voice_credentials
from rasa.core.channels.voice_stream.voice_channel import (
    ContinueConversationAction,
    EndConversationAction,
    NewAudioAction,
    VoiceChannelAction,
    VoiceInputChannel,
    VoiceOutputChannel,
)

logger = structlog.get_logger()


class BrowserAudioOutputChannel(VoiceOutputChannel):
    @classmethod
    def name(cls) -> str:
        return "browser_audio"

    def rasa_audio_bytes_to_channel_bytes(
        self, rasa_audio_bytes: RasaAudioBytes
    ) -> bytes:
        return audioop.ulaw2lin(rasa_audio_bytes, 4)

    def channel_bytes_to_message(self, recipient_id: str, channel_bytes: bytes) -> str:
        return json.dumps({"audio": base64.b64encode(channel_bytes).decode("utf-8")})

    def create_marker_message(self, recipient_id: str) -> Tuple[str, str]:
        message_id = uuid.uuid4().hex
        marker_data = {"marker": message_id}

        # Include comprehensive latency information if available
        latency_data = {
            "asr_latency_ms": call_state.asr_latency_ms,
            "rasa_processing_latency_ms": call_state.rasa_processing_latency_ms,
            "tts_first_byte_latency_ms": call_state.tts_first_byte_latency_ms,
            "tts_complete_latency_ms": call_state.tts_complete_latency_ms,
        }

        # Filter out None values from latency data
        latency_data = {k: v for k, v in latency_data.items() if v is not None}

        # Add latency data to marker if any metrics are available
        if latency_data:
            marker_data["latency"] = latency_data  # type: ignore[assignment]

        return json.dumps(marker_data), message_id


class BrowserAudioInputChannel(VoiceInputChannel):
    def __init__(
        self, server_url: str, asr_config: Dict[str, Any], tts_config: Dict[str, Any]
    ) -> None:
        """Initializes the browser audio input channel."""
        super().__init__(server_url, asr_config, tts_config)

    @classmethod
    def name(cls) -> str:
        return "browser_audio"

    def channel_bytes_to_rasa_audio_bytes(self, input_bytes: bytes) -> RasaAudioBytes:
        return RasaAudioBytes(audioop.lin2ulaw(input_bytes, 4))

    async def collect_call_parameters(
        self, channel_websocket: Websocket
    ) -> Optional[CallParameters]:
        call_id = f"inspect-{uuid.uuid4()}"
        return CallParameters(call_id, "local", "local", stream_id=call_id)

    @classmethod
    def from_credentials(
        cls,
        credentials: Optional[Dict[str, Any]],
    ) -> BrowserAudioInputChannel:
        cls.validate_basic_credentials(credentials)
        new_creds = repack_voice_credentials(credentials)
        return cls(**new_creds)

    def map_input_message(
        self,
        message: Any,
        ws: Websocket,
    ) -> VoiceChannelAction:
        data = json.loads(message)
        if "audio" in data:
            channel_bytes = base64.b64decode(data["audio"])
            audio_bytes = self.channel_bytes_to_rasa_audio_bytes(channel_bytes)
            return NewAudioAction(audio_bytes)
        elif "marker" in data:
            if data["marker"] == call_state.latest_bot_audio_id:
                # Just finished streaming last audio bytes
                call_state.is_bot_speaking = False
                if call_state.should_hangup:
                    logger.debug(
                        "browser_audio.hangup", marker=call_state.latest_bot_audio_id
                    )
                    return EndConversationAction()
            else:
                call_state.is_bot_speaking = True
        return ContinueConversationAction()

    def create_output_channel(
        self, voice_websocket: Websocket, tts_engine: TTSEngine
    ) -> VoiceOutputChannel:
        return BrowserAudioOutputChannel(
            voice_websocket,
            tts_engine,
            self.tts_cache,
        )

    def blueprint(
        self, on_new_message: Callable[[UserMessage], Awaitable[Any]]
    ) -> Blueprint:
        """Defines a Sanic blueprint"""
        blueprint = Blueprint("browser_audio", __name__)

        @blueprint.route("/", methods=["GET"])
        async def health(_: Request) -> HTTPResponse:
            return response.json({"status": "ok"})

        @blueprint.websocket("/websocket")  # type: ignore
        async def handle_message(request: Request, ws: Websocket) -> None:
            try:
                await self.run_audio_streaming(on_new_message, ws)
            except Exception as e:
                logger.error("browser_audio.handle_message.error", error=e)

        return blueprint



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict

from rasa.dialogue_understanding.stack.frames import PatternFlowStackFrame
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX

FLOW_PATTERN_SKIP_QUESTION = RASA_DEFAULT_FLOW_PATTERN_PREFIX + "skip_question"


@dataclass
class SkipQuestionPatternFlowStackFrame(PatternFlowStackFrame):
    """A pattern flow stack frame that gets added if user interrupts the current flow by
    trying to bypass the collect information step.
    """

    flow_id: str = FLOW_PATTERN_SKIP_QUESTION
    """The ID of the flow."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return FLOW_PATTERN_SKIP_QUESTION

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> SkipQuestionPatternFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return SkipQuestionPatternFlowStackFrame(
            frame_id=data["frame_id"],
            step_id=data["step_id"],
        )



from __future__ import annotations

from typing import Any, Dict, List, Optional

import structlog

from rasa.core.actions.action import Action
from rasa.core.channels import OutputChannel
from rasa.core.nlg import NaturalLanguageGenerator
from rasa.dialogue_understanding.stack.dialogue_stack import DialogueStack
from rasa.dialogue_understanding.stack.frames import (
    BaseFlowStackFrame,
    UserFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import FlowStackFrameType
from rasa.shared.core.constants import ACTION_CLEAN_STACK
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import Event
from rasa.shared.core.flows.steps.constants import END_STEP
from rasa.shared.core.flows.steps.continuation import ContinueFlowStep
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


class ActionCleanStack(Action):
    """Action which cancels a flow from the stack."""

    def name(self) -> str:
        """Return the flow name."""
        return ACTION_CLEAN_STACK

    async def run(
        self,
        output_channel: OutputChannel,
        nlg: NaturalLanguageGenerator,
        tracker: DialogueStateTracker,
        domain: Domain,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> List[Event]:
        """Clean the stack."""
        structlogger.debug("action_clean_stack.run")
        new_frames = []
        # Set all frames to their end step, filter out any non-BaseFlowStackFrames
        for frame in tracker.stack.frames:
            if isinstance(frame, BaseFlowStackFrame):
                structlogger.debug(
                    "action_clean_stack.terminating_frame",
                    frame_id=frame.frame_id,
                    flow_id=frame.flow_id,
                )
                frame.step_id = ContinueFlowStep.continue_step_for_id(END_STEP)
                if isinstance(frame, UserFlowStackFrame):
                    # Making sure there are no "continue interrupts" triggered
                    frame.frame_type = FlowStackFrameType.REGULAR
                new_frames.append(frame)
        new_stack = DialogueStack.from_dict([frame.as_dict() for frame in new_frames])

        return tracker.create_stack_updated_events(new_stack)
